/*
 *  Copyright (C) 2005  Thadeu Lima de Souza Cascardo <cascardo@holoscopio.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */



#include "lr0.h"
#include <stdlib.h>

transition_t* transition_new (gint action, gint state)
{
  transition_t* transition;
  transition = g_malloc (sizeof (transition_t));
  transition->action = action;
  transition->state = state;
  return transition;
}

static void lr0_push (lr0_t* parser, gint st, gpointer attrib)
{
  state_t* state;
  state = g_malloc (sizeof (state_t));
  state->state = st;
  state->attrib = attrib;
  parser->stack = g_list_prepend (parser->stack, state);
}

static gboolean lr0_pop (lr0_t* parser, gpointer* attrib)
{

  GList* l;
  state_t* state;
  if ((l = g_list_first (parser->stack)) == NULL)
    return FALSE;
  parser->stack = g_list_remove_link (l, l);
  state = (state_t*) l->data;
  if (attrib)
    *attrib = state->attrib;
  g_free (state);
  g_list_free (l);
  return TRUE;

}

lr0_t* lr0_new (nextcb cb, gpointer data)
{

  lr0_t* parser;

  parser = g_malloc (sizeof (lr0_t));
  parser->cb = cb;
  parser->data = data;

  parser->stack = NULL;
  parser_push (parser, 0, NULL);
  parser->table = g_hash_table_new_full (g_direct_hash, g_direct_equal,
					 NULL, g_hash_table_destroy);
  parser->rules = NULL;

  return parser;

}

void lr0_delete (lr0_t* parser)
{

  GList* l;

  for (l = g_list_first (parser->stack); l != NULL; l = g_list_next (l))
    {
      g_free (l->data);
    }

  g_list_free (parser->stack);

  g_hash_table_destroy (parser->table);

  g_free (parser);

}

void lr0_add (lr0_t* parser, gint state, symbol_t* symbol,
		 transition_t* transition)
{

  GHashTable* table;

  if (!g_hash_table_lookup_extended (parser->table, state,
				     NULL, (gpointer*) &table))
    {
      table = g_hash_table_new_full (symbol_hash, symbol_equal,
				     g_free, g_free);
      g_hash_table_insert (parser->table, state, table);
    }

  g_hash_table_insert (table, symbol, transition);

}

gboolean lr0_lookup (lr0_t* parser, gint state, symbol_t* symbol,
			transition_t** transition)
{

  GHashTable* table;
  transition_t* trans;

  if (!g_hash_table_lookup_extended (parser->table, state,
				     NULL, (gpointer*) &table))
    {
      return FALSE;
    }

  if (!g_hash_table_lookup_extended (table, symbol,
				     NULL, (gpointer*) &trans))
    {
      return FALSE;
    }

  if (transition)
    *transition = trans;

  return TRUE;

}

gpointer leaf_new (gpointer data)
{
  return g_node_new (data);
}

gpointer tree_new (rule_t* rule)
{
  return g_node_new (rule);
}

gpointer tree_add (gpointer tree, gpointer data)
{
  return g_node_prepend (tree, data);
}

gpointer lr0_build (lr0_t* parser)
{

  state_t* state;
  symbol_t* symbol;
  transition_t* transition;
  gpointer attrib;
  GList* l;

  symbol = g_malloc (sizeof (symbol_t));

  symbol->value = parser->cb (parser->data, &attrib);
  symbol->terminal = FALSE;

  while (1)
    {

      l = g_list_first (parser->stack);
      state = (state_t*) l->data;
      if (!lr0_lookup (parser, state->state, symbol, &transition))
	return NULL;

      if (transition->action == PARSER_SHIFT)
	{
	  lr0_push (parser, transition->state, leaf_new (attrib));
	  symbol->value = parser->cb (parser->data, &attrib);
	  symbol->terminal = FALSE;
	}

      else if (transition->action == PARSER_REDUCE)
	{

	  rule_t* rule;
	  state_t* state;
	  transition_t* trans;
	  GList* l;
	  gpointer attrib;

	  rule = g_list_nth_data (parser->rules, transition->state);
	  attrib = tree_new (rule);

	  for (l = g_list_last (rule->right);
	       l != NULL;
	       l = g_list_previous (l))
	    {
	      gpointer attr;
	      if (!lr0_pop (parser, &attr))
		return NULL;
	      tree_add (attrib, attr);
	    }

	  l = g_list_first (parser->stack);
	  state = (state_t*) l->data;
	  lr0_lookup (parser, state->state, rule->left, &trans);
	  lr0_push (parser, trans->state, attrib);

	}

      else if (transition->action == PARSER_ACCEPT)
	{
	  l = g_list_first (parser->stack);
	  state = (state_t*) l->data;
	  return state->attrib;
	}

    }

  return NULL;

}
